# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import dataclasses
import threading
import typing as tp
from contextlib import contextmanager
from types import MappingProxyType

A = tp.TypeVar('A')


@dataclasses.dataclass
class FlagsContext(threading.local):
  flags_stack: tp.List[MappingProxyType[str, tp.Any]] = dataclasses.field(
    default_factory=lambda: [MappingProxyType({})]
  )


FLAGS_CONTEXT = FlagsContext()


class Flags(tp.Mapping[str, tp.Any]):
  __slots__ = ()

  def __getitem__(self, name: str) -> tp.Any:
    current_flags = FLAGS_CONTEXT.flags_stack[-1]
    if name not in current_flags:
      raise ValueError(f'Unknown Flag: {name}')
    return current_flags[name]

  __getattr__ = __getitem__

  def __iter__(self) -> tp.Iterator[str]:
    return iter(FLAGS_CONTEXT.flags_stack[-1])

  def __len__(self) -> int:
    return len(FLAGS_CONTEXT.flags_stack[-1])

  def __contains__(self, name: tp.Any) -> bool:
    return name in FLAGS_CONTEXT.flags_stack[-1]

  @contextmanager
  def __call__(self, **kwargs: tp.Any):
    current_flags = FLAGS_CONTEXT.flags_stack[-1]
    FLAGS_CONTEXT.flags_stack.append(
      MappingProxyType(dict(current_flags, **kwargs))
    )
    try:
      yield
    finally:
      FLAGS_CONTEXT.flags_stack.pop()

  @tp.overload
  def get(self, name: str, /) -> tp.Any:
    ...

  @tp.overload
  def get(self, name: str, /, default: A) -> A:
    ...

  def get(self, name: str, /, default: A = None) -> A | None:
    return FLAGS_CONTEXT.flags_stack[-1].get(name, default)


flags = Flags()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import jax
import jax.numpy as jnp
import pytest

from flax.experimental import nnx
from flax.experimental.nnx.nnx.rnglib import _stable_hash


class TestRngs:
  def test_hash(self):
    _hash = _stable_hash('hi')
    assert isinstance(_hash, int)

  def test_call(self):
    rngs = nnx.Rngs(0)
    key = rngs()

  def test_fallback(self):
    rngs = nnx.Rngs(0)
    key = rngs.dropout()

  def test_fallback_error_no_default(self):
    rngs = nnx.Rngs(some_name=0)
    with pytest.raises(AttributeError, match='No RNG named'):
      key = rngs.dropout()

  def test_rng_stream(self):
    key0 = jax.random.key(0)
    rngs = nnx.Rngs(params=key0)
    assert rngs._rngs['params'].counts[-1] == 0

    key1 = rngs.params()
    assert rngs._rngs['params'].counts[-1] == 1
    assert rngs._rngs['params'].key is key0
    assert not jnp.allclose(key0, key1)

    key2 = rngs.params()
    assert rngs._rngs['params'].counts[-1] == 2
    assert rngs._rngs['params'].key is key0
    assert not jnp.allclose(key1, key2)

  def test_rng_fork(self):
    key0 = jax.random.key(0)
    rngs1 = nnx.Rngs(params=key0)
    rngs2 = nnx.Rngs(rngs1.fork())

    assert rngs2._rngs['params'].counts == [0, 0]

    key1 = rngs1.params()
    key2 = rngs2.params()

    assert not jnp.allclose(key1, key2)

  def test_rng_trace_level_constraints(self):
    rngs = nnx.Rngs(0)

    @jax.jit
    def f():
      with pytest.raises(
        nnx.TraceContextError,
        match='Cannot use Rngs from a different trace level',
      ):
        rngs.params()

    f()

    @jax.jit
    def f():
      with pytest.raises(
        nnx.TraceContextError,
        match='Cannot use Rngs from a different trace level',
      ):
        rngs.fork()

    f()

    rngs1: Any = None

    @jax.jit
    def g():
      nonlocal rngs1
      rngs1 = nnx.Rngs(1)

    g()

    assert isinstance(rngs1, nnx.Rngs)
    with pytest.raises(
      nnx.TraceContextError,
      match='Cannot use Rngs from a different trace level',
    ):
      rngs1.params()

  def test_partition_merge(self):
    rngs = nnx.Rngs(dropout=0)

    keys = rngs.fork()

    assert 'dropout' in keys
    assert keys['dropout'].counts == [0, 0]

    rngs2 = nnx.Rngs(keys)

    key1 = rngs.dropout()
    key2 = rngs2.dropout()
    assert not jnp.allclose(key1, key2)

    rngs3 = nnx.Rngs(keys)
    key3 = rngs3.dropout()
    assert jnp.allclose(key2, key3)

  def test_fork_broadcast(self):
    rngs = nnx.Rngs(params=0, dropout=1)
    jax.random.key

    keys = rngs.fork()  # all broadcast

    assert keys['params'].key.shape == ()
    assert keys['dropout'].key.shape == ()
    assert jnp.allclose(keys['params'].key, jax.random.key(0))
    assert jnp.allclose(keys['dropout'].key, jax.random.key(1))

  def test_fork_split(self):
    rngs = nnx.Rngs(params=0, dropout=1)
    keys = rngs.fork(4)  # split all

    assert keys['params'].key.shape == (4,)
    assert keys['dropout'].key.shape == (4,)

  def test_fork_split_and_broadcast(self):
    rngs = nnx.Rngs(params=0, dropout=1)
    splits, broadcasts = rngs.fork(params=4, dropout=None)

    assert splits['params'].key.shape == (4,)
    assert broadcasts['dropout'].key.shape == ()

  def test_fork_filters(self):
    rngs = nnx.Rngs(params=0, dropout=1)
    splits, broadcasts = rngs.fork({'params': 4})

    assert splits['params'].key.shape == (4,)
    assert broadcasts['dropout'].key.shape == ()

  def test_fork_multidimensional_split(self):
    rngs = nnx.Rngs(params=0, dropout=1)
    keys = rngs.fork((4, None, 3))  # split all

    assert keys['params'].key.shape == (4, 1, 3)
    assert keys['dropout'].key.shape == (4, 1, 3)

  def test_fork_multidimensional_split_mixed(self):
    rngs = nnx.Rngs(params=0, dropout=1)
    splits, broadcasts = rngs.fork(params=(4, None, 3))  # split all

    assert splits['params'].key.shape == (4, 1, 3)
    assert broadcasts['dropout'].key.shape == ()

  def test_rng_stream_pytree(self):
    rngs = nnx.Rngs(params=0, dropout=1)
    stream = rngs.fork()['params']

    stream2 = jax.tree_map(lambda x: x, stream)

    assert stream.key is stream2.key



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A variable dict is a normal Python dictionary, which is a container for one
or more "variable collections", each of which are nested dictionaries whose
leaves are ``jax.numpy`` arrays.

The different variable collections share the same nested tree structure.

For example, consider the following variable dictionary::

  {
    "params": {
      "Conv1": { "weight": ..., "bias": ... },
      "BatchNorm1": { "scale": ..., "mean": ... },
      "Conv2": {...}
    },
    "batch_stats": {
      "BatchNorm1": { "moving_mean": ..., "moving_average": ...}
    }
  }

In this case, the ``"BatchNorm1"`` key lives in both the ``"params"`` and
```"batch_stats""`` collections. This reflects the fact that the submodule
named ``""BatchNorm1""`` has both trainable parameters (the ``"params"`` collection),
as well as other non-trainable variables (the ``"batch_stats"`` collection)

TODO: Make "variable dict" design note, and link to it from here.
"""

from .scope import Variable



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for handling the Atari environment."""

import collections

import gymnasium as gym
import numpy as np

import seed_rl_atari_preprocessing


class ClipRewardEnv(gym.RewardWrapper):
  """Adapted from OpenAI baselines.

  github.com/openai/baselines/blob/master/baselines/common/atari_wrappers.py
  """

  def __init__(self, env):
    gym.RewardWrapper.__init__(self, env)

  def reward(self, reward):
    """Bin reward to {+1, 0, -1} by its sign."""
    return np.sign(reward)


class FrameStack:
  """Implements stacking of `num_frames` last frames of the game.

  Wraps an AtariPreprocessing object.
  """

  def __init__(
      self,
      preproc: seed_rl_atari_preprocessing.AtariPreprocessing,
      num_frames: int,
  ):
    self.preproc = preproc
    self.num_frames = num_frames
    self.frames = collections.deque(maxlen=num_frames)

  def reset(self):
    ob = self.preproc.reset()
    for _ in range(self.num_frames):
      self.frames.append(ob)
    return self._get_array()

  def step(self, action: int):
    ob, reward, done, info = self.preproc.step(action)
    self.frames.append(ob)
    return self._get_array(), reward, done, info

  def _get_array(self):
    assert len(self.frames) == self.num_frames
    return np.concatenate(self.frames, axis=-1)


def create_env(game: str, clip_rewards: bool):
  """Create a FrameStack object that serves as environment for the `game`."""
  env = gym.make(game)
  if clip_rewards:
    env = ClipRewardEnv(env)  # bin rewards to {-1., 0., 1.}
  preproc = seed_rl_atari_preprocessing.AtariPreprocessing(env)
  stack = FrameStack(preproc, num_frames=4)
  return stack


def get_num_actions(game: str):
  """Get the number of possible actions of a given Atari game.

  This determines the number of outputs in the actor part of the
  actor-critic model.
  """
  env = gym.make(game)
  return env.action_space.n



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main file for running the SST2 example.
This file is intentionally kept short. The majority for logic is in libraries
that can be easily tested and imported in Colab.
"""

from absl import app
from absl import flags
from absl import logging
from clu import platform
import jax
from ml_collections import config_flags
import tensorflow as tf

import train


FLAGS = flags.FLAGS

flags.DEFINE_string('workdir', None, 'Directory to store model data.')
config_flags.DEFINE_config_file(
    'config',
    None,
    'File path to the training hyperparameter configuration.',
    lock_config=True,
)
flags.mark_flags_as_required(['config', 'workdir'])


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  # Hide any GPUs from TensorFlow. Otherwise TF might reserve memory and make
  # it unavailable to JAX.
  tf.config.experimental.set_visible_devices([], 'GPU')

  logging.info('JAX process: %d / %d', jax.process_index(), jax.process_count())
  logging.info('JAX local devices: %r', jax.local_devices())

  # Add a note so that we can tell which task is which JAX host.
  # (Depending on the platform task 0 is not guaranteed to be host 0)
  platform.work_unit().set_task_status(
      f'process_index: {jax.process_index()}, '
      f'process_count: {jax.process_count()}'
  )
  platform.work_unit().create_artifact(
      platform.ArtifactType.DIRECTORY, FLAGS.workdir, 'workdir'
  )

  train.train_and_evaluate(FLAGS.config, FLAGS.workdir)


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Input pipeline for seq2seq addition example."""

import random
from typing import Any, Dict, Generator, Optional, Tuple

import jax.numpy as jnp
import numpy as np

Array = Any  # pylint: disable=invalid-name


class CharacterTable:
  """Encodes/decodes between strings and integer representations."""

  def __init__(self, chars: str, max_len_query_digit: int = 3) -> None:
    self._chars = sorted(set(chars))
    self._char_indices = {ch: idx + 2 for idx, ch in enumerate(self._chars)}
    self._indices_char = {idx + 2: ch for idx, ch in enumerate(self._chars)}
    self._indices_char[self.pad_id] = '_'
    # Maximum length of a single input digit.
    self._max_len_query_digit = max_len_query_digit

  @property
  def pad_id(self) -> int:
    return 0

  @property
  def eos_id(self) -> int:
    return 1

  @property
  def vocab_size(self) -> int:
    # All characters + pad token and eos token.
    return len(self._chars) + 2

  @property
  def max_input_len(self) -> int:
    """Returns the max length of an input sequence."""
    # The input has the form "digit1+digit2<eos>", so the max input length is
    # the length of two digits plus two tokens for "+" and the EOS token.
    return self._max_len_query_digit * 2 + 2

  @property
  def max_output_len(self) -> int:
    """Returns the max length of an output sequence."""
    # The output has the form "=digit<eos>". If `digit` is the result of adding
    # two digits of max length x, then max length of `digit` is x+1.
    # Additionally, we require two more tokens for "=" and "<eos".
    return self._max_len_query_digit + 3

  @property
  def encoder_input_shape(self) -> Tuple[int, int, int]:
    return (1, self.max_input_len, self.vocab_size)

  @property
  def decoder_input_shape(self) -> Tuple[int, int, int]:
    return (1, self.max_output_len, self.vocab_size)

  def encode(self, inputs: str) -> np.ndarray:
    """Encodes from string to list of integers."""
    return np.array(
        [self._char_indices[char] for char in inputs] + [self.eos_id]
    )

  def decode(self, inputs: Array) -> str:
    """Decodes from list of integers to string."""
    chars = []
    for elem in inputs.tolist():
      if elem == self.eos_id:
        break
      chars.append(self._indices_char[elem])
    return ''.join(chars)

  def one_hot(self, tokens: np.ndarray) -> np.ndarray:
    vecs = np.zeros((tokens.size, self.vocab_size), dtype=np.float32)
    vecs[np.arange(tokens.size), tokens] = 1
    return vecs

  def encode_onehot(
      self, batch_inputs: Array, max_len: Optional[int] = None
  ) -> np.ndarray:
    """One-hot encodes a string input."""

    if max_len is None:
      max_len = self.max_input_len

    def encode_str(s):
      tokens = self.encode(s)
      unpadded_len = len(tokens)
      if unpadded_len > max_len:
        raise ValueError(f"Sequence too long ({len(tokens)}>{max_len}): '{s}'")
      tokens = np.pad(tokens, [(0, max_len - len(tokens))], mode='constant')
      return self.one_hot(tokens)

    return np.array([encode_str(inp) for inp in batch_inputs])

  def decode_onehot(self, batch_inputs: Array) -> np.ndarray:
    """Decodes a batch of one-hot encoding to strings."""
    decode_inputs = lambda inputs: self.decode(inputs.argmax(axis=-1))
    return np.array(list(map(decode_inputs, batch_inputs)))

  def generate_examples(
      self, num_examples: int
  ) -> Generator[Tuple[str, str], None, None]:
    """Yields `num_examples` examples."""
    for _ in range(num_examples):
      max_digit = pow(10, self._max_len_query_digit) - 1
      # TODO(marcvanzee): Use jax.random here.
      key = tuple(sorted((random.randint(0, 99), random.randint(0, max_digit))))
      inputs = f'{key[0]}+{key[1]}'
      # Preprend output by the decoder's start token.
      outputs = '=' + str(key[0] + key[1])
      yield (inputs, outputs)

  def get_batch(self, batch_size: int) -> Dict[str, np.ndarray]:
    """Returns a batch of example of size @batch_size."""
    inputs, outputs = zip(*self.generate_examples(batch_size))
    return {
        'query': self.encode_onehot(inputs),
        'answer': self.encode_onehot(outputs),
    }


def mask_sequences(sequence_batch: Array, lengths: Array) -> Array:
  """Sets positions beyond the length of each sequence to 0."""
  return sequence_batch * (
      lengths[:, np.newaxis] > np.arange(sequence_batch.shape[1])[np.newaxis]
  )


def get_sequence_lengths(sequence_batch: Array, eos_id: int) -> Array:
  """Returns the length of each one-hot sequence, including the EOS token."""
  # sequence_batch.shape = (batch_size, seq_length, vocab_size)
  eos_row = sequence_batch[:, :, eos_id]
  eos_idx = jnp.argmax(eos_row, axis=-1)  # returns first occurrence
  # `eos_idx` is 0 if EOS is not present, so we use full length in that case.
  return jnp.where(
      eos_row[jnp.arange(eos_row.shape[0]), eos_idx],
      eos_idx + 1,
      sequence_batch.shape[1],  # if there is no EOS, use full length
  )



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Script for creating VM on Google cloud and run a Flax example inside.
# See ./README.md for instructions.

import datetime
import os
import re
import subprocess
import time
from typing import Sequence

from absl import app
from absl import flags

# General options.
flags.DEFINE_bool(
    'dry_run',
    False,
    help=(
        'If set, then the command to launch the GCE instance will only be '
        'printed to stdout.'
    ),
)
flags.DEFINE_bool(
    'connect',
    False,
    help='Same as --wait, but directly connect to VM once it is ready.',
)
flags.DEFINE_bool(
    'wait',
    False,
    help=(
        'If set, then the script will wait until VM is ready. If VM_READY_CMD '
        'is set in environment, then that command will be executed once the VM '
        'is ready. Useful for sending a notification, e.g. "osascript" (mac).'
    ),
)

# Machine configuration.
flags.DEFINE_string('project', None, help='Name of the Google Cloud project.')
flags.DEFINE_string('zone', None, help='Zone in which the VM will be created.')
flags.DEFINE_string(
    'machine_type',
    None,
    help='Machine type to use for VM. See "gcloud compute machine-types list".',
)
flags.DEFINE_string(
    'accelerator_type',
    '',
    help=(
        'Type of accelerator to use, or empty. '
        'See "gcloud compute accelerator-types list".'
    ),
)
flags.DEFINE_integer(
    'shutdown_secs',
    300,
    help=(
        'How long to wait (after successful/failed training) before shutting '
        'down the VM. Set to 0 to disable.'
    ),
)
flags.DEFINE_integer(
    'accelerator_count', 8, help='Number of accelerators to use.'
)

# GCS configuration.
flags.DEFINE_string(
    'gcs_workdir_base',
    None,
    help=(
        'GCS base directory for model output. The --workdir argument will be '
        'constructed from {gcs_workdir_base}/{example}/{name}/{timestamp} .'
    ),
)
flags.DEFINE_string(
    'tfds_data_dir',
    '',
    help=(
        'Optional tfds data directory. This can be useful to prepare datasets'
        ' on GCS and then point the jobs to this preloaded directory. Dataset'
        ' will be downloaded from the web if not specified.'
    ),
)

# Repo configuration.
flags.DEFINE_string(
    'repo', 'https://github.com/google/flax', help='Git repository'
)
flags.DEFINE_string('branch', 'main', help='Git repository')

# Example configuration.
flags.DEFINE_string(
    'example', None, help='Name of Flax example (e.g. "imagenet").'
)
flags.DEFINE_string(
    'args',
    '',
    help=(
        'Any additional command line arguments for {example}_main.py, like '
        'for example --config. Note that --workdir will be provided by the '
        'script.'
    ),
)

# Run configuration.
flags.DEFINE_string(
    'name',
    None,
    help=(
        'Name of the experiment. Note that the provided name will be '
        'extended to {example}/{name}/{timestamp}'
    ),
)

FLAGS = flags.FLAGS
flags.mark_flags_as_required(
    ['project', 'zone', 'machine_type', 'gcs_workdir_base', 'example', 'name']
)

timestamp = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')


def generate_startup_file(vm_name: str) -> str:
  directory = os.path.dirname(os.path.abspath(__file__))
  startup_script_src = os.path.join(directory, 'startup_script.sh')
  startup_script_dst = os.path.join(directory, f'{vm_name}-startup_script.sh')
  assert not os.path.exists(startup_script_dst)
  with open(startup_script_src, encoding='utf8') as f:
    startup_script_content = f.read()
  for from_str, to_str in (
      ('__REPO__', FLAGS.repo),
      ('__BRANCH__', FLAGS.branch),
      ('__EXAMPLE__', FLAGS.example),
      ('__TIMESTAMP__', timestamp),
      ('__NAME__', FLAGS.name),
      ('__ARGS__', FLAGS.args),
      ('__GCS_WORKDIR_BASE__', FLAGS.gcs_workdir_base),
      ('__TFDS_DATA_DIR__', FLAGS.tfds_data_dir),
      ('__ACCELERATOR_TYPE__', FLAGS.accelerator_type),
      ('__SHUTDOWN_SECS__', str(FLAGS.shutdown_secs)),
  ):
    startup_script_content = startup_script_content.replace(from_str, to_str)
  with open(startup_script_dst, 'w', encoding='utf8') as f:
    f.write(startup_script_content)
  return startup_script_dst


def launch_gce(*, vm_name: str, startup_script: str):
  # Note : Use `gcloud compute images list --project ml-images` to get a list
  # of available VM images.
  args = [
      'gcloud',
      'compute',
      'instances',
      'create',
      vm_name,
      f'--project={FLAGS.project}',
      f'--zone={FLAGS.zone}',
      '--image=c1-deeplearning-tf-2-10-cu113-v20221107-debian-10',
      '--image-project=ml-images',
      f'--machine-type={FLAGS.machine_type}',
      '--scopes=cloud-platform,storage-full',
      '--boot-disk-size=256GB',
      '--boot-disk-type=pd-ssd',
      '--metadata=install-nvidia-driver=True',
      f'--metadata-from-file=startup-script={startup_script}',
  ]
  if FLAGS.accelerator_type and FLAGS.accelerator_count:
    args.extend([
        '--maintenance-policy=TERMINATE',
        f'--accelerator=type={FLAGS.accelerator_type},count={FLAGS.accelerator_count}',
    ])

  if FLAGS.dry_run:
    print()
    print('Would run the following command without --dry-run:')
    print()
    print(' \\\n    '.join(args))
    print()
    return

  print()
  print('Creating instance on GCE... This will take some minutes...')
  print()
  result = subprocess.run(args)
  if result.returncode:
    raise RuntimeError('Could not create VM!')


def print_howto(login_args: Sequence[str]):
  print(f"""
###############################################################################
###############################################################################

You can start/stop the instace via the web UI:
https://console.cloud.google.com/compute/instances?project={FLAGS.project}

Once the VM has started, you can login and connect to the training session:

{' '.join(login_args)}

Note that you can disconnect from the tmux session without stopping the training
with the keystrokes 'CTRL-B A'. See "man tmux" for help about tmux.

To observe the training via Tensorboard, simply run in your local computer:

$ tensorboard --logdir={FLAGS.gcs_workdir_base}

You can also browse the files at

https://console.cloud.google.com/storage/browser/{FLAGS.gcs_workdir_base.replace('gs://', '')}

###############################################################################
###############################################################################
""")


def main(_):
  for name in ('repo', 'branch', 'example', 'name', 'gcs_workdir_base'):
    value = getattr(FLAGS, name)
    if re.match(r'[^\w:/_-]', value):
      raise ValueError(f'Invalid flag value: --{name}="{value}"')
  example_base_directory = os.path.join(
      os.path.dirname(os.path.abspath(__file__)),
      os.path.pardir,
  )
  if not os.path.isdir(os.path.join(example_base_directory, FLAGS.example)):
    raise ValueError(f'Could not find --example={FLAGS.example}')
  if FLAGS.connect and FLAGS.dry_run:
    raise ValueError('Cannot --connect to VM with --dry_run')

  vm_name = '-'.join([
      'flax',
      FLAGS.example,
      timestamp,
  ])
  vm_name = re.sub(r'[^a-z0-9-]', '-', vm_name)

  startup_script = generate_startup_file(vm_name)
  launch_gce(vm_name=vm_name, startup_script=startup_script)

  login_args = [
      'gcloud',
      'compute',
      'ssh',
      '--project',
      FLAGS.project,
      '--zone',
      FLAGS.zone,
      vm_name,
      '--',
      '/sudo_tmux_a.sh',
  ]

  print('Your instance is being started...')

  print_howto(login_args)

  if FLAGS.connect or FLAGS.wait:
    login_true_args = login_args[:-1] + ['true']
    while True:
      try:
        result = subprocess.run(
            login_true_args, timeout=10, stderr=subprocess.PIPE
        )
        if result.returncode == 0:
          break
        stderr = result.stderr.decode('utf8')
        if 'connection refused' in stderr.lower():
          print('(Connection refused - waiting a little longer...)')
          time.sleep(20)
        else:
          raise ValueError(f'Unknown error: {stderr}')
      except ValueError as e:
        if 'HTTP 502' not in str(e):
          raise e
        print('(Bad Gateway - waiting a little longer...)')
        time.sleep(20)
      except subprocess.TimeoutExpired:
        print('(Timeout - waiting a little longer...)')
        time.sleep(20)
    if 'VM_READY_CMD' in os.environ:
      os.system(os.environ['VM_READY_CMD'])
    if FLAGS.connect:
      result = subprocess.run(login_args)
      # SSH session has cleared previous message, print it again.
      print_howto(login_args)


if __name__ == '__main__':
  app.run(main)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Callable, Optional, Sequence

import jax
from absl.testing import absltest
from jax import lax, random
from jax import numpy as jnp

from flax.core import Array, Scope, init, lift, nn, unfreeze


def softmax_attn(scope: Scope, weights: Array):
  del scope
  norm_dims = tuple(range(weights.ndim // 2, weights.ndim))
  log_norms = jax.scipy.special.logsumexp(
    weights, axis=norm_dims, keepdims=True
  )
  return jnp.exp(weights - log_norms)


def with_dropout(fn, rate: float, deterministic: bool = False):
  def attn_fn(scope: Scope, weights: Array):
    attn_weights = fn(scope, weights)
    return nn.dropout(
      scope, attn_weights, deterministic=deterministic, rate=rate
    )

  return attn_fn


def _dot_product_attention(
  scope: Scope,
  query: Array,
  key: Array,
  value: Array,
  bias: Optional[Array] = None,
  attn_fn: Callable = softmax_attn,
  dtype=jnp.float32,
):
  assert key.ndim == query.ndim
  assert key.ndim == value.ndim

  n = query.ndim
  attn_weights = lax.dot_general(query, key, (((n - 1,), (n - 1,)), ((), ())))
  if bias is not None:
    attn_weights += bias
  attn_weights = attn_fn(scope, attn_weights)
  attn_weights = attn_weights.astype(dtype)

  contract_dims = (
    tuple(range(n - 1, attn_weights.ndim)),
    tuple(range(0, n - 1)),
  )
  y = lax.dot_general(attn_weights, value, (contract_dims, ((), ())))
  return y


def dot_product_attention(
  scope: Scope,
  inputs_q: Array,
  inputs_kv: Array,
  bias: Optional[Array] = None,
  qkv_features: Optional[int] = None,
  out_features: Optional[int] = None,
  attn_fn: Callable = softmax_attn,
  dtype=jnp.float32,
):
  if qkv_features is None:
    qkv_features = inputs_q.shape[-1]
  if out_features is None:
    out_features = inputs_q.shape[-1]
  dense = partial(nn.dense, features=qkv_features, bias=False, dtype=dtype)

  query = scope.child(dense, 'query')(inputs_q)
  key = scope.child(dense, 'key')(inputs_kv)
  value = scope.child(dense, 'value')(inputs_kv)

  y = _dot_product_attention(
    scope, query, key, value, bias=bias, attn_fn=attn_fn, dtype=dtype
  )

  return scope.child(nn.dense, 'out')(y, features=out_features, dtype=dtype)


def multi_head_dot_product_attention(
  scope: Scope,
  inputs_q: Array,
  inputs_kv: Array,
  bias: Optional[Array] = None,
  qkv_features: Optional[int] = None,
  out_features: Optional[int] = None,
  attn_fn: Callable = softmax_attn,
  batch_axes: Sequence[int] = (0,),
  num_heads: int = 1,
  dtype=jnp.float32,
  broadcast_dropout=False,
):
  if qkv_features is None:
    qkv_features = inputs_q.shape[-1]
  if out_features is None:
    out_features = inputs_q.shape[-1]

  attn_fn = partial(
    dot_product_attention,
    attn_fn=attn_fn,
    qkv_features=qkv_features // num_heads,
    out_features=out_features,
    dtype=dtype,
  )
  attn_fn = lift.vmap(
    attn_fn,
    in_axes=(None, None, None),
    out_axes=-2,
    axis_size=num_heads,
    variable_axes={'params': 0},
    split_rngs={'params': True, 'dropout': not broadcast_dropout},
  )
  for axis in reversed(sorted(batch_axes)):
    attn_fn = lift.vmap(
      attn_fn,
      in_axes=(axis, axis, axis),
      out_axes=axis,
      variable_axes={'params': None},
      split_rngs={'params': False, 'dropout': not broadcast_dropout},
    )

  y = attn_fn(scope, inputs_q, inputs_kv, bias)
  return y.mean(axis=-2)


class AttentionTest(absltest.TestCase):
  def test_attention(self):
    inputs = jnp.ones((2, 7, 16))
    model = partial(
      multi_head_dot_product_attention,
      num_heads=2,
      batch_axes=(0,),
      attn_fn=with_dropout(softmax_attn, 0.1, deterministic=False),
    )

    rngs = {'params': random.key(0), 'dropout': random.key(1)}
    y, variables = jax.jit(init(model))(rngs, inputs, inputs)
    variable_shapes = jax.tree_util.tree_map(jnp.shape, variables['params'])
    self.assertEqual(y.shape, (2, 7, 16))
    self.assertEqual(
      unfreeze(variable_shapes),
      {
        'key': {'kernel': (2, 16, 8)},
        'value': {'kernel': (2, 16, 8)},
        'query': {'kernel': (2, 16, 8)},
        'out': {'bias': (2, 16), 'kernel': (2, 8, 16)},
      },
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as tp

import jax
import jax.numpy as jnp
from jax import random

from flax.experimental import nnx


class Linear(nnx.Module):
  @tp.overload
  def __init__(self, *, din: int, dout: int, rngs: nnx.Rngs):
    ...

  @tp.overload
  def __init__(self, *, dout: int):
    ...

  @tp.overload
  def __init__(
    self,
    *,
    din: tp.Optional[int] = None,
    dout: int,
    rngs: tp.Optional[nnx.Rngs] = None,
  ):
    ...

  def __init__(
    self,
    *,
    din: tp.Optional[int] = None,
    dout: int,
    rngs: tp.Optional[nnx.Rngs] = None,
  ):
    self.dout = dout
    if din is not None:
      if rngs is None:
        raise ValueError('rngs must be provided if din is provided')
      self.init_variables(din, rngs)

  def init_variables(self, din: int, rngs: nnx.Rngs):
    key = rngs.params()
    self.w = nnx.Param(random.uniform(key, (din, self.dout)))
    self.b = nnx.Param(jnp.zeros((self.dout,)))

  def __call__(
    self, x: jax.Array, *, rngs: tp.Optional[nnx.Rngs] = None
  ) -> jax.Array:
    if self.is_initializing and not hasattr(self, 'w'):
      if rngs is None:
        raise ValueError('rngs must be provided to initialize module')
      self.init_variables(x.shape[-1], rngs)

    return x @ self.w + self.b


class BatchNorm(nnx.Module):
  @tp.overload
  def __init__(self, *, mu: float = 0.95):
    ...

  @tp.overload
  def __init__(self, *, din: int, mu: float = 0.95, rngs: nnx.Rngs):
    ...

  @tp.overload
  def __init__(
    self,
    *,
    din: tp.Optional[int] = None,
    mu: float = 0.95,
    rngs: tp.Optional[nnx.Rngs] = None,
  ):
    ...

  def __init__(
    self,
    *,
    din: tp.Optional[int] = None,
    mu: float = 0.95,
    rngs: tp.Optional[nnx.Rngs] = None,
  ):
    self.mu = mu

    if din is not None:
      if rngs is None:
        raise ValueError('rngs must be provided if din is provided')
      self.init_variables(din, rngs)

  def init_variables(self, din: int, rngs: nnx.Rngs):
    self.scale = nnx.Param(jax.numpy.ones((din,)))
    self.bias = nnx.Param(jax.numpy.zeros((din,)))
    self.mean = nnx.BatchStat(jax.numpy.zeros((din,)))
    self.var = nnx.BatchStat(jax.numpy.ones((din,)))

  def __call__(
    self, x, *, train: bool, rngs: tp.Optional[nnx.Rngs] = None
  ) -> jax.Array:
    if self.is_initializing and not hasattr(self, 'scale'):
      if rngs is None:
        raise ValueError('rngs must be provided to initialize module')
      self.init_variables(x.shape[-1], rngs)

    if train:
      axis = tuple(range(x.ndim - 1))
      mean = jax.numpy.mean(x, axis=axis)
      var = jax.numpy.var(x, axis=axis)
      # ema update
      self.mean = self.mu * self.mean + (1 - self.mu) * mean
      self.var = self.mu * self.var + (1 - self.mu) * var
    else:
      mean, var = self.mean, self.var

    scale, bias = self.scale, self.bias
    x = (x - mean) / jax.numpy.sqrt(var + 1e-5) * scale + bias
    return x


class Dropout(nnx.Module):
  def __init__(self, rate: float):
    self.rate = rate

  def __call__(self, x: jax.Array, *, train: bool, rngs: nnx.Rngs) -> jax.Array:
    if train:
      mask = random.bernoulli(rngs.dropout(), (1 - self.rate), x.shape)
      x = x * mask / (1 - self.rate)
    return x


# ----------------------------
# test Linear
# ----------------------------
print('test Linear')

# eager
m1 = Linear(din=32, dout=10, rngs=nnx.Rngs(params=0))
y = m1(x=jnp.ones((1, 32)))
print(jax.tree_map(jnp.shape, m1.get_state()))

# lazy
m2 = Linear(dout=10)
y = m2.init(x=jnp.ones((1, 32)), rngs=nnx.Rngs(params=0))
print(jax.tree_map(jnp.shape, m2.get_state()))

# usage
y1 = m1(x=jnp.ones((1, 32)))
y2 = m2(x=jnp.ones((1, 32)))

# ----------------------------
# Test scan
# ----------------------------
print('\ntest scan')


class Block(nnx.Module):
  def __init__(
    self,
    din: tp.Optional[int] = None,
    dout: int = 10,
    rngs: tp.Optional[nnx.Rngs] = None,
  ):
    self.linear = Linear(din=din, dout=dout, rngs=rngs)
    self.bn = BatchNorm(din=dout if din is not None else None, rngs=rngs)
    self.dropout = Dropout(0.5)

  def __call__(self, x: jax.Array, _, *, train: bool, rngs: nnx.Rngs):
    x = self.linear(x, rngs=rngs)
    x = self.bn(x, train=train, rngs=rngs)
    x = self.dropout(x, train=train, rngs=rngs)
    x = jax.nn.gelu(x)
    return x, None


MLP = nnx.Scan(
  Block,
  variable_axes={nnx.Param: 0},
  variable_carry=nnx.BatchStat,
  split_rngs={'params': True, 'dropout': True},
  length=5,
)


# eager
mlp = MLP(din=10, dout=10, rngs=nnx.Rngs(params=0))
y, _ = mlp.call(jnp.ones((1, 10)), None, train=True, rngs=nnx.Rngs(dropout=1))
print(f'{y.shape=}')
print('state =', jax.tree_map(jnp.shape, mlp.get_state()))
print()

# lazy
mlp = MLP(dout=10)
mlp.init(jnp.ones((1, 10)), None, train=False, rngs=nnx.Rngs(params=0))
y, _ = mlp.call(jnp.ones((1, 10)), None, train=True, rngs=nnx.Rngs(dropout=1))
print(f'{y.shape=}')
print('state =', jax.tree_map(jnp.shape, mlp.get_state()))



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from flax.linen.pooling import avg_pool as avg_pool
from flax.linen.pooling import max_pool as max_pool
from flax.linen.pooling import min_pool as min_pool
from flax.linen.pooling import pool as pool

from .nnx import compatibility as compatibility
from .nnx.dataclasses import dataclass as dataclass
from .nnx.dataclasses import field as field
from .nnx import graph_utils as graph_utils
from .nnx.dataclasses import param_field as param_field
from .nnx.dataclasses import treenode_field as treenode_field
from .nnx.dataclasses import variable_field as variable_field
from .nnx.errors import TraceContextError as TraceContextError
from .nnx.filterlib import All as All
from .nnx.filterlib import Not as Not
from .nnx.flaglib import flags as flags
from .nnx.graph_utils import GraphDef as GraphDef
from .nnx.helpers import Dict as Dict
from .nnx.helpers import Sequence as Sequence
from .nnx.helpers import TrainState as TrainState
from .nnx.module import GraphDef as GraphDef
from .nnx.module import M as M
from .nnx.module import Module as Module
from .nnx.module import merge as merge
from .nnx.nn import initializers as initializers
from .nnx.nn.activations import celu as celu
from .nnx.nn.activations import elu as elu
from .nnx.nn.activations import gelu as gelu
from .nnx.nn.activations import glu as glu
from .nnx.nn.activations import hard_sigmoid as hard_sigmoid
from .nnx.nn.activations import hard_silu as hard_silu
from .nnx.nn.activations import hard_swish as hard_swish
from .nnx.nn.activations import hard_tanh as hard_tanh
from .nnx.nn.activations import leaky_relu as leaky_relu
from .nnx.nn.activations import log_sigmoid as log_sigmoid
from .nnx.nn.activations import log_softmax as log_softmax
from .nnx.nn.activations import logsumexp as logsumexp
from .nnx.nn.activations import one_hot as one_hot
from .nnx.nn.activations import relu as relu
from .nnx.nn.activations import relu6 as relu6
from .nnx.nn.activations import selu as selu
from .nnx.nn.activations import sigmoid as sigmoid
from .nnx.nn.activations import silu as silu
from .nnx.nn.activations import soft_sign as soft_sign
from .nnx.nn.activations import softmax as softmax
from .nnx.nn.activations import softplus as softplus
from .nnx.nn.activations import standardize as standardize
from .nnx.nn.activations import swish as swish
from .nnx.nn.activations import tanh as tanh
from .nnx.nn.attention import MultiHeadAttention as MultiHeadAttention
from .nnx.nn.attention import combine_masks as combine_masks
from .nnx.nn.attention import dot_product_attention as dot_product_attention
from .nnx.nn.attention import make_attention_mask as make_attention_mask
from .nnx.nn.attention import make_causal_mask as make_causal_mask
from .nnx.nn.initializers import Initializer as Initializer
from .nnx.nn.linear import Conv as Conv
from .nnx.nn.linear import Embed as Embed
from .nnx.nn.linear import Linear as Linear
from .nnx.nn.linear import LinearGeneral as LinearGeneral
from .nnx.nn.normalization import BatchNorm as BatchNorm
from .nnx.nn.normalization import LayerNorm as LayerNorm
from .nnx.nn.stochastic import Dropout as Dropout
from .nnx.pytreelib import Pytree as Pytree
from .nnx.pytreelib import TreeNode as TreeNode
from .nnx.rnglib import Rngs as Rngs
from .nnx.rnglib import RngStream as RngStream
from .nnx.spmd import PARTITION_NAME as PARTITION_NAME
from .nnx.spmd import get_partition_spec as get_partition_spec
from .nnx.spmd import get_named_sharding as get_named_sharding
from .nnx.spmd import with_partitioning as with_partitioning
from .nnx.spmd import with_sharding_constraint as with_sharding_constraint
from .nnx.state import State as State
from .nnx.transforms import JIT as JIT
from .nnx.transforms import Remat as Remat
from .nnx.transforms import Scan as Scan
from .nnx.transforms import Vmap as Vmap
from .nnx.transforms import grad as grad
from .nnx.transforms import jit as jit
from .nnx.transforms import remat as remat
from .nnx.transforms import scan as scan
from .nnx.transforms import value_and_grad as value_and_grad
from .nnx.transforms import vmap as vmap
from .nnx.variables import EMPTY as EMPTY
from .nnx.variables import A as A
from .nnx.variables import BatchStat as BatchStat
from .nnx.variables import Cache as Cache
from .nnx.variables import Empty as Empty
from .nnx.variables import Intermediate as Intermediate
from .nnx.variables import Param as Param
from .nnx.variables import Rng as Rng
from .nnx.variables import Variable as Variable
from .nnx.variables import VariableMetadata as VariableMetadata
from .nnx.variables import with_metadata as with_metadata



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import dataclasses
import enum
import typing as tp

import jax

from flax.experimental.nnx.nnx import filterlib, reprlib
from flax.experimental.nnx.nnx.proxy_caller import (
  ApplyCaller,
  CallableProxy,
  DelayedAccessor,
)
from flax.experimental.nnx.nnx.state import State
from flax.experimental.nnx.nnx.variables import EMPTY, Empty, Variable

HA = tp.TypeVar('HA', bound=tp.Hashable)
HB = tp.TypeVar('HB', bound=tp.Hashable)

Index = int
Names = tp.Sequence[int]
PathParts = tuple[str, ...]
Path = str
Node = tp.TypeVar('Node')
Leaf = tp.TypeVar('Leaf')
AuxData = tp.TypeVar('AuxData')

NODE_TYPES: dict[type, 'NodeImpl[tp.Any, tp.Any, tp.Any]'] = {}


@dataclasses.dataclass(frozen=True)
class NodeImpl(tp.Generic[Node, Leaf, AuxData]):
  type: type
  flatten: tp.Callable[[Node], tuple[tp.Sequence[tuple[str, Leaf]], AuxData]]
  get_key: tp.Callable[[Node, str], Leaf]
  set_key: tp.Callable[[Node, str, Leaf], Node]
  has_key: tp.Callable[[Node, str], bool]
  all_keys: tp.Callable[[Node], tuple[str, ...]]
  unflatten: tp.Callable[[tuple[tuple[str, Leaf], ...], AuxData], Node] | None
  create_empty: tp.Callable[[AuxData], Node] | None
  init: tp.Callable[[Node, tuple[tuple[str, Leaf], ...]], None] | None

  def items(self, node: Node) -> tp.Iterator[tuple[str, Leaf]]:
    for key in self.all_keys(node):
      yield key, self.get_key(node, key)


@tp.overload
def register_node_type(
  type: type,
  flatten: tp.Callable[[Node], tuple[tp.Sequence[tuple[str, Leaf]], AuxData]],
  get_key: tp.Callable[[Node, str], Leaf],
  set_key: tp.Callable[[Node, str, Leaf], Node],
  has_key: tp.Callable[[Node, str], bool],
  all_keys: tp.Callable[[Node], tuple[str, ...]],
  *,
  unflatten: tp.Callable[[tuple[tuple[str, Leaf], ...], AuxData], Node],
):
  ...


@tp.overload
def register_node_type(
  type: type,
  flatten: tp.Callable[[Node], tuple[tp.Sequence[tuple[str, Leaf]], AuxData]],
  get_key: tp.Callable[[Node, str], Leaf],
  set_key: tp.Callable[[Node, str, Leaf], Node],
  has_key: tp.Callable[[Node, str], bool],
  all_keys: tp.Callable[[Node], tuple[str, ...]],
  *,
  create_empty: tp.Callable[[AuxData], Node],
  init: tp.Callable[[Node, tuple[tuple[str, Leaf], ...]], None],
):
  ...


def register_node_type(
  type: type,
  flatten: tp.Callable[[Node], tuple[tp.Sequence[tuple[str, Leaf]], AuxData]],
  get_key: tp.Callable[[Node, str], Leaf],
  set_key: tp.Callable[[Node, str, Leaf], Node],
  has_key: tp.Callable[[Node, str], bool],
  all_keys: tp.Callable[[Node], tuple[str, ...]],
  *,
  unflatten: tp.Callable[[tuple[tuple[str, Leaf], ...], AuxData], Node]
  | None = None,
  create_empty: tp.Callable[[AuxData], Node] | None = None,
  init: tp.Callable[[Node, tuple[tuple[str, Leaf], ...]], None] | None = None,
):
  if type in NODE_TYPES:
    raise ValueError(f"Node type '{type}' already registered.")
  NODE_TYPES[type] = NodeImpl(
    type,
    flatten,
    get_key,
    set_key,
    has_key,
    all_keys,
    unflatten,
    create_empty,
    init,
  )


def is_node(x: tp.Any) -> bool:
  return type(x) in NODE_TYPES


def is_node_type(x: type[tp.Any]) -> bool:
  return x in NODE_TYPES


@tp.overload
def get_node_impl(x: type[Node]) -> NodeImpl[Node, tp.Any, tp.Any]:
  ...


@tp.overload
def get_node_impl(x: Node) -> NodeImpl[Node, tp.Any, tp.Any]:
  ...


def get_node_impl(x: type[Node] | Node) -> NodeImpl[Node, tp.Any, tp.Any]:
  if not isinstance(x, type):
    x = type(x)
  if not is_node_type(x):
    raise ValueError(f'Unknown node type: {x}')
  return NODE_TYPES[x]


class _HashableMapping(tp.Mapping[HA, HB], tp.Hashable):
  def __init__(self, mapping: tp.Mapping[HA, HB] | tp.Iterable[tuple[HA, HB]]):
    self._mapping = dict(mapping)

  def __contains__(self, key: object) -> bool:
    return key in self._mapping

  def __getitem__(self, key: HA) -> HB:
    return self._mapping[key]

  def __iter__(self) -> tp.Iterator[HA]:
    return iter(self._mapping)

  def __len__(self) -> int:
    return len(self._mapping)

  def __hash__(self) -> int:
    return hash(tuple(self._mapping.items()))

  def __eq__(self, other: tp.Any) -> bool:
    if not isinstance(other, tp.Mapping):
      return False
    return self._mapping == other

  def __repr__(self) -> str:
    return repr(self._mapping)


@dataclasses.dataclass(repr=False)
class _MappingRepr(reprlib.Representable):
  mapping: tp.Mapping[str, tp.Any]

  def __nnx_repr__(self):
    yield reprlib.Object(type='', value_sep=': ', start='{', end='}')

    for key, value in self.mapping.items():
      yield reprlib.Attr(repr(key), value)


class VariableDef(reprlib.Representable):
  __slots__ = (
    '_type',
    '_index',
    '_metadata',
  )

  @classmethod
  def from_variable(cls, variable: Variable[tp.Any], index: int) -> VariableDef:
    metadata = vars(variable).copy()
    del metadata['value']
    return cls(type(variable), index, metadata)

  def to_variable(self, value: Node) -> Variable[Node]:
    variables = object.__new__(self._type)
    variables.value = value
    vars(variables).update(self._metadata)
    return variables

  def __init__(
    self,
    type: tp.Type[Variable[tp.Any]],
    index: int,
    metadata: dict[str, tp.Any],
  ):
    self._type = type
    self._index = index
    self._metadata = metadata

  def __nnx_repr__(self):
    yield reprlib.Object(type=type(self))

    yield reprlib.Attr('type', self._type.__name__)
    yield reprlib.Attr('index', self._index)
    yield reprlib.Attr('metadata', _MappingRepr(self._metadata))

  @property
  def type(self):
    return self._type

  @property
  def index(self):
    return self._index

  @property
  def metadata(self):
    return self._metadata

  def __hash__(self):
    return hash((self._type, self._index, tuple(self._metadata.items())))

  def __eq__(self, other):
    if not isinstance(other, VariableDef):
      return False
    return (
      self._type == other._type
      and self._index == other._index
      and self._metadata == other._metadata
    )


class GraphDef(tp.Generic[Node], reprlib.Representable):
  __slots__ = (
    '_type',
    '_index',
    '_attributes',
    '_subgraphs',
    '_static_fields',
    '_variables',
    '_metadata',
  )

  def __init__(
    self,
    type: tp.Type[Node],
    index: int,
    attributes: tuple[str, ...],
    subgraphs: tp.Iterable[tuple[str, tp.Union['GraphDef[tp.Any]', int]]],
    static_fields: tp.Iterable[tuple[str, tp.Any]],
    variables: tp.Iterable[tuple[str, VariableDef | int]],
    metadata: tp.Any,
  ):
    self._type = type
    self._index = index
    self._attributes = attributes
    self._subgraphs = _HashableMapping(subgraphs)
    self._static_fields = _HashableMapping(static_fields)
    self._variables = _HashableMapping(variables)
    self._metadata = metadata

  def __nnx_repr__(self):
    yield reprlib.Object(type=type(self))

    yield reprlib.Attr('type', self._type.__name__)
    yield reprlib.Attr('index', self._index)
    yield reprlib.Attr('attributes', self._attributes)
    yield reprlib.Attr('subgraphs', _MappingRepr(self._subgraphs))
    yield reprlib.Attr('static_fields', _MappingRepr(self._static_fields))
    yield reprlib.Attr('variables', _MappingRepr(self._variables))
    yield reprlib.Attr('metadata', self._metadata)

  def __hash__(self) -> int:
    return hash((self._type, self._subgraphs))

  def __eq__(self, other: tp.Any) -> bool:
    if not isinstance(other, GraphDef):
      return False
    return self._type == other._type and self._subgraphs == other._subgraphs

  @property
  def type(self) -> tp.Type[Node]:
    return self._type

  @property
  def index(self) -> int:
    return self._index

  @property
  def attributes(self) -> tuple[str, ...]:
    return self._attributes

  @property
  def subgraphs(self):
    return self._subgraphs

  @property
  def static_fields(self):
    return self._static_fields

  @property
  def variables(self):
    return self._variables

  @property
  def metadata(self) -> tp.Any:
    return self._metadata

  def merge(self, state: State, *states: State) -> Node:
    if states:
      state = State.merge(state, *states)
    return graph_unflatten(self, state)

  def apply(
    self, state: State, *states: State
  ) -> ApplyCaller[tuple[State, 'GraphDef[Node]']]:
    accessor = DelayedAccessor()

    def _apply(
      accessor: DelayedAccessor, *args, **kwargs
    ) -> tuple[tp.Any, tuple[State, GraphDef[Node]]]:
      module = self.merge(state, *states)
      fn = accessor(module)
      out = fn(*args, **kwargs)
      return out, graph_flatten(module)

    return CallableProxy(_apply, accessor)  # type: ignore

  def make_empty(self) -> Node:
    return self.merge(State({}))


def _gradphdef_flatten(graphdef: GraphDef[tp.Any]):
  return (), (
    graphdef._type,
    graphdef._index,
    graphdef._attributes,
    graphdef._subgraphs,
    graphdef._static_fields,
    graphdef._variables,
    graphdef._metadata,
  )


def _graphdef_unflatten(
  metadata: tuple[
    tp.Type[Node],
    int,
    tuple[str, ...],
    tuple[tuple[str, GraphDef[Node] | int], ...],
    tuple[tuple[str, tp.Any], ...],
    tuple[tuple[str, Variable[Empty] | int], ...],
    tp.Any,
  ],
  _,
) -> GraphDef[Node]:
  return GraphDef(*metadata)


jax.tree_util.register_pytree_node(
  GraphDef, _gradphdef_flatten, _graphdef_unflatten
)


def graph_flatten(x: Node) -> tuple[State, GraphDef[Node]]:
  id_to_index: dict[int, Index] = {}
  flat_state: dict[Path, Variable[tp.Any]] = {}
  dagdef = _graph_flatten((), id_to_index, flat_state, x)
  assert not isinstance(dagdef, int)
  return State.from_flat_path(flat_state), dagdef


def _graph_flatten(
  path: PathParts,
  id_to_index: dict[int, Index],
  flat_state: dict[Path, Variable[tp.Any]],
  node: Node,
) -> GraphDef[Node] | int:
  if not is_node(node):
    raise RuntimeError(f'Unsupported type: {type(node)}, this is a bug.')

  if (node_id := id(node)) in id_to_index:
    return id_to_index[node_id]

  index = len(id_to_index)
  id_to_index[id(node)] = index

  subgraphs: list[tuple[str, tp.Union[GraphDef[Node], int]]] = []
  static_fields: list[tuple[str, tp.Any]] = []
  variables: list[tuple[str, VariableDef | int]] = []

  node_impl = get_node_impl(node)
  values, metadata = node_impl.flatten(node)
  for key, value in values:
    if not isinstance(key, str):
      raise TypeError(
        f'Node (of type {type(node).__name__}) has a key of non-string '
        f'type {type(key).__name__}.'
      )
    if is_node(value):
      graphdef = _graph_flatten((*path, key), id_to_index, flat_state, value)
      subgraphs.append((key, graphdef))
    elif isinstance(value, Variable):
      str_path = '/'.join((*path, key))
      variable_id = id(value)
      if variable_id in id_to_index:
        variables.append((key, id_to_index[variable_id]))
      else:
        flat_state[str_path] = value.copy()
        variable_index = id_to_index[variable_id] = len(id_to_index)
        variables.append(
          (key, VariableDef.from_variable(value, variable_index))
        )
    else:
      static_fields.append((key, value))

  graphdef = GraphDef(
    type=type(node),
    index=index,
    attributes=tuple(key for key, _ in values),
    subgraphs=subgraphs,
    static_fields=static_fields,
    variables=variables,
    metadata=metadata,
  )
  return graphdef


def graph_unflatten(graphdef: GraphDef[Node], state: State) -> Node:
  index_to_node: dict[Index, tp.Any] = {}
  return _graph_unflatten(graphdef, state.raw_mapping, index_to_node)


def _graph_unflatten(
  graphdef: tp.Union[GraphDef[Node], int],
  state: dict[str, Variable[tp.Any] | dict[str, tp.Any]],
  index_to_node: dict[Index, tp.Any],
) -> Node:
  if isinstance(graphdef, int):
    return index_to_node[graphdef]

  if not is_node_type(graphdef.type):
    raise RuntimeError(f'Unsupported type: {graphdef.type}, this is a bug.')

  if graphdef.index in index_to_node:
    raise RuntimeError(f'GraphDef index {graphdef.index} already used.')

  state = state.copy()
  node_impl = get_node_impl(graphdef.type)

  def _get_children():
    new_state: dict[str, tp.Any] = {}

    for key in graphdef.attributes:
      if key in graphdef.static_fields:
        new_state[key] = graphdef.static_fields[key]
      elif key not in state:
        # if key is not present create an empty types
        if key in graphdef.subgraphs:
          # if the key is a subgraph we create an empty node
          subgraphdef = graphdef.subgraphs[key]
          if isinstance(subgraphdef, int):
            # subgraph exists, take it from the cache
            new_state[key] = index_to_node[subgraphdef]
          else:
            # create an empty node and add it to the cache
            substate = {}
            node = new_state[key] = _graph_unflatten(
              subgraphdef, substate, index_to_node
            )
            index_to_node[subgraphdef.index] = node
        elif key in graphdef.variables:
          variable_def = graphdef.variables[key]
          if isinstance(variable_def, int):
            # variable exists, take it from the cache
            new_state[key] = index_to_node[variable_def]
          else:
            # create an empty variable and add it to the cache
            node = new_state[key] = variable_def.to_variable(EMPTY)
            index_to_node[variable_def.index] = node
        else:
          raise RuntimeError(f'Unknown static field: {key!r}')
      else:
        value = state[key]
        if key in graphdef.subgraphs:
          if isinstance(value, Variable):
            raise ValueError(
              f'Expected a subgraph for {key!r}, but got a Variable.'
            )
          subgraphdef = graphdef.subgraphs[key]

          if isinstance(subgraphdef, int):
            node = index_to_node[subgraphdef]
          else:
            node = new_state[key] = _graph_unflatten(
              subgraphdef, value, index_to_node
            )
            index_to_node[subgraphdef.index] = node

        elif key in graphdef.variables:
          variable_def = graphdef.variables[key]
          if isinstance(variable_def, int):
            new_state[key] = index_to_node[variable_def]
          else:
            if type(value) != variable_def.type:
              raise ValueError(
                f'Expected a Variable of type {variable_def.type} '
                f'for {key!r}, but got a Variable of type {type(value)}.'
              )
            assert isinstance(value, Variable)
            value = value.copy()
            new_state[key] = value
            index_to_node[variable_def.index] = value

    for new_key in set(state) - set(graphdef.attributes):
      new_state[new_key] = state[new_key]

    return new_state

  if node_impl.create_empty:
    assert node_impl.init is not None
    # we create an empty node first and add it to the index
    # this avoids infinite recursion when there is a reference cycle
    node = node_impl.create_empty(graphdef.metadata)
    index_to_node[graphdef.index] = node
    children = _get_children()
    node_impl.init(node, tuple(children.items()))
  else:
    # if the node type does not support the creation of an empty object it means
    # that it cannot reference itself, so we can create its children first
    assert node_impl.unflatten is not None
    children = _get_children()
    node = node_impl.unflatten(tuple(children.items()), graphdef.metadata)
    index_to_node[graphdef.index] = node

  return node


def graph_pop(
  node: tp.Any,
  filters: tuple[filterlib.Filter, ...],
) -> tuple[State, ...]:
  id_to_index: dict[int, Index] = {}
  path_parts: PathParts = ()
  predicates = tuple(filterlib.to_predicate(filter) for filter in filters)
  states = tuple({} for _ in predicates)
  _graph_pop(node, id_to_index, path_parts, states, predicates)
  return tuple(State(x) for x in states)


def _graph_pop(
  node: tp.Any,
  id_to_index: dict[int, Index],
  path_parts: PathParts,
  states: tuple[dict[Path, tp.Any], ...],
  predicates: tuple[filterlib.Predicate, ...],
) -> None:
  if not is_node(node):
    raise RuntimeError(f'Unsupported type: {type(node)}, this is a bug.')

  if id(node) in id_to_index:
    return

  id_to_index[id(node)] = len(id_to_index)
  node_impl = get_node_impl(node)

  for name in node_impl.all_keys(node):
    value = node_impl.get_key(node, name)
    if is_node(value):
      _graph_pop(value, id_to_index, (*path_parts, name), states, predicates)
      continue
    elif not isinstance(value, Variable):
      continue
    elif value.is_empty:
      continue
    elif id(value) in id_to_index:
      continue

    path = '/'.join((*path_parts, name))
    node_impl = get_node_impl(node)
    for state, predicate in zip(states, predicates):
      if predicate(path, value):
        state[path] = value.copy()
        id_to_index[id(value)] = len(id_to_index)
        # empty Variable attributes
        node_impl.set_key(node, name, value.as_empty())
        break
    else:
      # NOTE: should we raise an error here?
      pass


def graph_update_dynamic(
  node: tp.Any,
  updates: State | tp.Sequence[State],
) -> None:
  if not is_node(node):
    raise ValueError(f'Unsupported type: {type(node)}')

  if isinstance(updates, State):
    new_states = (updates,)
  else:
    new_states = updates

  for state in new_states:
    _graph_update_dynamic(node, state.raw_mapping)


def _graph_update_dynamic(
  node: tp.Any, state: dict[str, Variable[tp.Any] | dict[str, tp.Any]]
):
  if not is_node(node):
    raise RuntimeError(f'Unsupported type: {type(node)}')

  node_impl = get_node_impl(node)
  for key, value in state.items():
    if is_node(value):
      if isinstance(value, Variable):
        raise ValueError(
          f'Expected a subgraph for {key!r}, but got a Variable.'
        )
      _graph_update_dynamic(node_impl.get_key(node, key), value)
    else:
      # we skip if trying to replace a non-empty Variable with an empty one
      if (
        isinstance(value, Variable)
        and value.is_empty
        and isinstance(current_value := node_impl.get_key(node, key), Variable)
        and not current_value.is_empty
      ):
        continue

      node_impl.set_key(node, key, value)


class _StaticModuleStatus(enum.Enum):
  NEW = enum.auto()
  UPDATED = enum.auto()


def graph_update_static(node: Node, updates: Node) -> None:
  cache: dict[int, _StaticModuleStatus] = {}
  _graph_update_static(node, updates, cache, _StaticModuleStatus.UPDATED, ())


def _graph_update_static(
  node: Node,
  updates: Node,
  cache: dict[int, _StaticModuleStatus],
  status: _StaticModuleStatus,
  path: PathParts,
) -> None:
  if type(node) != type(updates):
    raise ValueError(
      f'Trying to update a node with a different type: '
      f'expected {type(node).__name__!r}, '
      f'but got {type(updates).__name__!r}'
    )
  if not is_node(node):
    raise ValueError(f'Unsupported node type: {type(node)}')

  if id(updates) in cache:
    if cache[id(updates)] != status:
      str_path = '/'.join(path)
      if status is _StaticModuleStatus.NEW:
        raise ValueError(
          f'Trying to add a new node at path {str_path!r} but a'
          ' node with the same reference has been updated'
        )
      else:
        raise ValueError(
          f'Trying to update a node at path {str_path!r} but a new'
          ' node with the same reference has been added'
        )
    return

  cache[id(updates)] = status

  node_impl = get_node_impl(node)
  for name, value_updates in node_impl.items(updates):
    if isinstance(value_updates, Variable):
      continue
    elif is_node(value_updates):
      if node_impl.has_key(node, name):
        _graph_update_static(
          node_impl.get_key(node, name),
          value_updates,
          cache,
          _StaticModuleStatus.UPDATED,
          (*path, name),
        )
      else:
        if id(value_updates) in cache:
          if cache[id(value_updates)] is not _StaticModuleStatus.NEW:
            raise ValueError(
              f'Trying to add a new node at path {name!r} but a '
              'node with the same reference has been updated'
            )
        else:
          cache[id(value_updates)] = _StaticModuleStatus.NEW

        node_impl.set_key(node, name, value_updates)
    else:  # static field
      node_impl.set_key(node, name, value_updates)


def clone(node: Node) -> Node:
  state, static = graph_flatten(node)
  return static.merge(state)


def iter_nodes(node: tp.Any) -> tp.Iterator[tuple[Path, tp.Any]]:
  visited: set[int] = set()
  path_parts: PathParts = ()
  yield from _iter_nodes(node, visited, path_parts)


def _iter_nodes(
  node: tp.Any, visited: set[int], path_parts: PathParts
) -> tp.Iterator[tuple[Path, tp.Any]]:
  if not is_node(node):
    return
  if id(node) in visited:
    return
  visited.add(id(node))
  path = '/'.join(path_parts)
  yield path, node
  node_impl = get_node_impl(node)
  for key, value in node_impl.items(node):
    yield from _iter_nodes(value, visited, (*path_parts, key))


# -----------------------------
# register node types
# -----------------------------
# dict
def _flatten_dict(
  node: dict[str, tp.Any],
) -> tuple[tuple[tuple[str, tp.Any], ...], None]:
  return tuple(node.items()), None


def _get_key_dict(node: dict[str, tp.Any], key: str) -> tp.Any:
  return node[key]


def _set_key_dict(
  node: dict[str, tp.Any], key: str, value: tp.Any
) -> dict[str, tp.Any]:
  node[key] = value
  return node


def _has_key_dict(node: dict[str, tp.Any], key: str) -> bool:
  return key in node


def _all_keys_dict(node: dict[str, tp.Any]) -> tuple[str, ...]:
  return tuple(node.keys())


def _create_empty_dict(metadata: None) -> dict[str, tp.Any]:
  return {}


def _init_dict(node: dict[str, tp.Any], items: tuple[tuple[str, tp.Any], ...]):
  node.update(items)


register_node_type(
  dict,
  _flatten_dict,
  _get_key_dict,
  _set_key_dict,
  _has_key_dict,
  _all_keys_dict,
  create_empty=_create_empty_dict,
  init=_init_dict,
)


# list
def _flatten_list(
  node: list[tp.Any],
) -> tuple[tuple[tuple[str, tp.Any], ...], int]:
  return tuple((str(i), value) for i, value in enumerate(node)), len(node)


def _get_key_list(node: list[tp.Any], key: str) -> tp.Any:
  return node[int(key)]


def _set_key_list(node: list[tp.Any], key: str, value: tp.Any) -> list[tp.Any]:
  int_key = int(key)
  if int_key >= len(node):
    node.extend([EMPTY] * (int_key - len(node) + 1))
  node[int_key] = value
  return node


def _has_key_list(node: list[tp.Any], key: str) -> bool:
  return int(key) < len(node)


def _all_keys_list(node: list[tp.Any]) -> tuple[str, ...]:
  return tuple(str(i) for i in range(len(node)))


def _create_empty_list(length: int) -> list[tp.Any]:
  return [EMPTY] * length


def _init_list(node: list[tp.Any], items: tuple[tuple[str, tp.Any], ...]):
  for key, value in items:
    _set_key_list(node, key, value)


register_node_type(
  list,
  _flatten_list,
  _get_key_list,
  _set_key_list,
  _has_key_list,
  _all_keys_list,
  create_empty=_create_empty_list,
  init=_init_list,
)


# tuple
def _flatten_tuple(
  node: tuple[tp.Any, ...],
) -> tuple[tuple[tuple[str, tp.Any], ...], int]:
  return tuple((str(i), value) for i, value in enumerate(node)), len(node)


def _unflatten_tuple(
  items: tuple[tuple[str, tp.Any], ...], length: int
) -> tuple[tp.Any, ...]:
  node = [EMPTY] * length
  for key, value in items:
    node[int(key)] = value
  return tuple(node)


def _get_key_tuple(node: tuple[tp.Any, ...], key: str) -> tp.Any:
  return node[int(key)]


def _set_key_tuple(node: tuple[tp.Any, ...], key: str, value: tp.Any):
  current_value = _get_key_tuple(node, key)
  if current_value is value or current_value == value:
    return
  else:
    raise ValueError("'tuple' object is immutable, does not support assignment")


def _has_key_tuple(node: tuple[tp.Any, ...], key: str) -> bool:
  return int(key) < len(node)


def _all_keys_tuple(node: tuple[tp.Any, ...]) -> tuple[str, ...]:
  return tuple(str(i) for i in range(len(node)))


register_node_type(
  tuple,
  _flatten_tuple,
  _get_key_tuple,
  _set_key_tuple,
  _has_key_tuple,
  _all_keys_tuple,
  unflatten=_unflatten_tuple,
)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import numpy as np
from absl.testing import absltest
from jax import numpy as jnp
from jax import random

from flax import errors
from flax.configurations import temp_flip_flag
from flax.core import Scope, apply, freeze, init, lazy_init, nn, scope
from flax.core.scope import LazyRng


class ScopeTest(absltest.TestCase):
  def test_rng(self):
    def f(scope):
      self.assertTrue(scope.has_rng('params'))
      self.assertFalse(scope.has_rng('dropout'))
      rng = scope.make_rng('params')
      self.assertTrue(
        np.all(rng == LazyRng.create(random.key(0), 1).as_jax_rng())
      )

    init(f)(random.key(0))

  def test_in_filter(self):
    filter_true = lambda x, y: self.assertTrue(scope.in_filter(x, y))
    filter_false = lambda x, y: self.assertFalse(scope.in_filter(x, y))

    filter_true(True, 'any_string1')
    filter_false(False, 'any_string2')
    filter_true('exact_match', 'exact_match')
    filter_false('no_match1', 'no_match2')
    filter_true(['one', 'two'], 'one')
    filter_false(['one', 'two'], 'three')
    filter_false([], 'one')
    filter_false([], None)

  def test_union_filter(self):
    def union_check(a, b, ans):
      self.assertEqual(scope.union_filters(a, b), ans)
      self.assertEqual(scope.union_filters(b, a), ans)

    union_check(['a', 'b'], ['b', 'c'], set(['a', 'b', 'c']))
    union_check(True, False, True)
    union_check(False, False, set())
    union_check(True, True, True)
    union_check(
      scope.DenyList(['a', 'b']),
      scope.DenyList(['b', 'c']),
      scope.DenyList(set(['b'])),
    )
    union_check(
      scope.DenyList(['a', 'b']), ['b', 'c'], scope.DenyList(set(['a']))
    )

  def test_intersect_filter(self):
    def intersect_check(a, b, ans):
      self.assertEqual(scope.intersect_filters(a, b), ans)
      self.assertEqual(scope.intersect_filters(b, a), ans)

    intersect_check(['a', 'b'], ['b', 'c'], set(['b']))
    intersect_check(True, False, False)
    intersect_check(False, False, set())
    intersect_check(True, True, True)
    intersect_check(
      scope.DenyList(['a', 'b']),
      scope.DenyList(['b', 'c']),
      scope.DenyList(set(['a', 'b', 'c'])),
    )
    intersect_check(scope.DenyList(['a', 'b']), ['b', 'c'], set(['c']))

  def test_subtract_filter(self):
    def subtract_check(a, b, ans):
      self.assertEqual(scope.subtract_filters(a, b), ans)

    subtract_check(['a', 'b'], ['b', 'c'], set(['a']))
    subtract_check(True, False, scope.DenyList(False))
    subtract_check(False, False, set())
    subtract_check(True, True, False)
    subtract_check(True, 'a', scope.DenyList('a'))
    subtract_check(
      scope.DenyList(['a', 'b']), scope.DenyList(['b', 'c']), set(['c'])
    )
    subtract_check(
      scope.DenyList(['a', 'b']),
      ['b', 'c'],
      scope.DenyList(set(['a', 'b', 'c'])),
    )

  def test_group_collections(self):
    params = {'dense1': {'x': [10, 20]}}
    batch_stats = {'dense1': {'ema': 5}}
    xs = {'params': params, 'batch_stats': batch_stats}

    # Retrieve all keys only once.
    group = scope.group_collections(xs, ['params', 'params'])
    self.assertEqual(group, ({'params': params}, {}))

    # Ignore non-existing keys.
    self.assertEqual(scope.group_collections(xs, ['vars']), ({},))

    # False gets nothing and True retrieves all keys once.
    self.assertEqual(
      scope.group_collections(xs, [False, True, True]), ({}, xs, {})
    )

  def test_inconsistent_param_shapes(self):
    def f(scope):
      scope.param('test', nn.initializers.ones_init(), (4,))

    msg = (
      r'Initializer expected to generate shape \(2,\) but got shape \(4,\)'
      r' instead for parameter "test" in "/"'
    )
    with self.assertRaisesRegex(errors.ScopeParamShapeError, msg):
      apply(f)(freeze({'params': {'test': np.ones((2,))}}))

  def test_apply_variables_bad_pytree(self):
    def f(scope):
      scope.param('kernel', nn.initializers.ones_init(), (4,))

    params = freeze(
      {
        'params': {
          'kernel': np.ones((4,)),
        },
      }
    )
    apply(f)(params)  # Valid.
    msg = 'but got a dict with an extra params layer'
    with self.assertRaisesRegex(
      errors.ApplyScopeInvalidVariablesStructureError, msg
    ):
      apply(f)({'params': params})

  def test_mutate_undefined_collection(self):
    def f(scope):
      scope.put_variable('state', 'test', 123)

    msg = (
      r'Cannot update variable "test" in "/" because collection "state" is'
      r' immutable.'
    )
    with self.assertRaisesRegex(errors.ModifyScopeVariableError, msg):
      init(f, mutable='params')(random.key(0))

  def test_undefined_param(self):
    def f(scope):
      nn.dense(scope.push('dense'), np.ones((1, 2)), 2)

    msg = r'Could not find parameter named "kernel" in scope "/dense".'
    with self.assertRaisesRegex(errors.ScopeParamNotFoundError, msg):
      apply(f)({'params': {'abc': 1}})

  def test_variable_is_mutable(self):
    def f(scope, should_be_mutable):
      test = scope.variable('state', 'test', lambda: 1)
      self.assertEqual(test.is_mutable(), should_be_mutable)

    _, variables = apply(f, mutable='state')({}, True)
    apply(f, mutable=False)(variables, False)

  def test_rngs_check_w_frozen_dict(self):
    def f(scope, x):
      return x

    _ = apply(f)({}, np.array([0.0]), rngs=freeze({'a': random.key(0)}))

  def test_rng_check_w_old_and_new_keys(self):
    # random.key always returns a new-style typed PRNG key.
    key = random.key(0)
    self.assertTrue(scope._is_valid_rng(key))
    self.assertFalse(scope._is_valid_rng(random.split(key)))

    # random.PRNGKey returns an old-style uint32 key by default.
    old_key = random.PRNGKey(0)
    self.assertTrue(scope._is_valid_rng(old_key))
    self.assertFalse(scope._is_valid_rng(random.split(old_key)))

    # Also explicitly test raw key data, because the jax_enable_custom_prng
    # flag can make PRNGKey return new-style keys.
    raw_key = random.key_data(key)
    self.assertTrue(scope._is_valid_rng(raw_key))
    self.assertFalse(scope._is_valid_rng(random.split(raw_key)))

  def test_jax_leak_detector(self):
    with jax.check_tracer_leaks(True):

      def f(scope):
        def g(scope):
          pass

        scope.child(g)()

      jax.jit(init(f))(random.key(0))

  def test_rng_counter_reuse(self):
    root = Scope({}, {'dropout': random.key(0)})

    def f(scope):
      return scope.make_rng('dropout')

    a = root.child(f)()
    root = root.rewound()
    b = root.child(f)()
    self.assertFalse(jnp.allclose(a, b))

  def test_empty_col_error(self):
    root = Scope({})
    with self.assertRaises(errors.ScopeCollectionNotFound):
      root.param('test', nn.initializers.zeros_init(), ())
    root = Scope({'params': {}})
    with self.assertRaises(errors.ScopeCollectionNotFound):
      root.param('test', nn.initializers.zeros_init(), ())

    root = Scope({'params': {'abc': 1}})
    with self.assertRaises(errors.ScopeCollectionNotFound):
      root.variable('state', 'test', jnp.zeros, ())
    root = Scope({'state': {}})
    with self.assertRaises(errors.ScopeCollectionNotFound):
      root.variable('state', 'test', jnp.zeros, ())

  def test_variable_no_init(self):
    root = Scope({}, mutable='state')
    with self.assertRaises(errors.ScopeCollectionNotFound):
      root.variable('state', 'test')
    root = Scope({'state': {'abc': 1}}, mutable='state')
    abc = root.variable('state', 'abc')
    self.assertEqual(abc.value, 1)
    with self.assertRaises(errors.ScopeVariableNotFoundError):
      root.variable('state', 'test')

  def test_variable_alias(self):
    scope = Scope({}, mutable='state')
    subscope = scope.push(name='a')
    subscope.put_variable('state', 'x', 0.0)
    scope.put_variable('state', 'a', {'x': jnp.array(1.0, jnp.float32)})
    self.assertEqual(
      scope.variables()['state']['a']['x'], subscope.variables()['state']['x']
    )

  def test_lazy_init(self):
    def f(scope, x):
      k = scope.param(
        'kernel', nn.initializers.lecun_normal(), (x.shape[-1], x.shape[-1])
      )
      return x @ k

    init_fn = lazy_init(f)
    # provide a massive input message which would OOM if any compute ops were actually executed
    variables = init_fn(
      random.key(0),
      jax.ShapeDtypeStruct((1024 * 1024 * 1024, 128), jnp.float32),
    )
    self.assertEqual(variables['params']['kernel'].shape, (128, 128))

  def test_lazy_init_fails_on_data_dependence(self):
    def f(scope, x):
      # kernel is initialized with x so params are now dependent on the input
      k = scope.param('kernel', lambda _: x)
      return x * k

    init_fn = lazy_init(f)
    with self.assertRaises(errors.LazyInitError):
      init_fn(random.key(0), jax.ShapeDtypeStruct((8, 4), jnp.float32))

  @temp_flip_flag('fix_rng_separator', True)
  def test_fold_in_static_seperator(self):
    x = LazyRng(random.key(0), ('ab', 'c'))
    y = LazyRng(random.key(0), ('a', 'bc'))
    self.assertFalse(np.all(x.as_jax_rng() == y.as_jax_rng()))


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import pytest

from flax.experimental import nnx


class TestGraphUtils:
  def test_flatten(self):
    a = {'a': 1, 'b': nnx.Param(2)}
    g = [a, 3, a, nnx.Param(4)]

    state, static = nnx.graph_utils.graph_flatten(g)

    state['0']['b'] = 2
    state['3'] = 4

  def test_unflatten(self):
    a = {'a': 1, 'b': nnx.Param(2)}
    g = [a, 3, a, nnx.Param(4)]

    state, static = nnx.graph_utils.graph_flatten(g)
    g = static.merge(state)

    assert g[0] is g[2]

  def test_unflatten_empty(self):
    a = {'a': 1, 'b': nnx.Param(2)}
    g = [a, 3, a, nnx.Param(4)]

    state, static = nnx.graph_utils.graph_flatten(g)
    g = static.merge(nnx.State({}))

    assert g[0] is g[2]
    assert g[0]['b'].value is nnx.EMPTY
    assert g[3].value is nnx.EMPTY

  def test_update_dynamic(self):
    a = {'a': 1, 'b': nnx.Param(2)}
    g = [a, 3, a, nnx.Param(4)]

    state, static = nnx.graph_utils.graph_flatten(g)

    state['0']['b'] = 3
    nnx.graph_utils.graph_update_dynamic(g, state)

    assert g[0]['b'].value == 3
    assert g[2]['b'].value == 3

  def test_update_static(self):
    a = {'a': 1, 'b': nnx.Param(2)}
    g = [a, 3, a, nnx.Param(4)]

    g2 = nnx.graph_utils.clone(g)
    g2[0]['a'] = 5

    nnx.graph_utils.graph_update_static(g, g2)

    assert g[0]['a'] == 5
    assert g[2]['a'] == 5

  def test_update_static_inconsistent_types(self):
    a = {'a': 1, 'b': nnx.Param(2)}
    g = [a, 3, a, nnx.Param(4)]
    g2 = [a, a, 3, nnx.Param(4)]

    with pytest.raises(
      ValueError, match='Trying to update a node with a different type'
    ):
      nnx.graph_utils.graph_update_static(g, g2)

  def test_update_static_add_new(self):
    a = {'a': 1, 'b': nnx.Param(2)}
    b = [5, 6]
    g = [a, 3, a, nnx.Param(4)]
    g2 = [a, 3, a, nnx.Param(4), b]

    nnx.graph_utils.graph_update_static(g, g2)

    assert g[4][0] == 5
    assert g[4][1] == 6

  def test_update_static_add_shared_error(self):
    a = {'a': 1, 'b': nnx.Param(2)}
    g = [a, 3, a, nnx.Param(4)]
    g2 = [a, 3, a, nnx.Param(4), a]

    with pytest.raises(ValueError, match='Trying to add a new node at path'):
      nnx.graph_utils.graph_update_static(g, g2)

  def test_module_list(self):
    rngs = nnx.Rngs(0)
    ls = [
      nnx.Linear(2, 2, rngs=rngs),
      nnx.BatchNorm(2, rngs=rngs),
    ]

    state, static = nnx.graph_utils.graph_flatten(ls)

    assert state['0']['kernel'].shape == (2, 2)
    assert state['0']['bias'].shape == (2,)
    assert state['1']['scale'].shape == (2,)
    assert state['1']['bias'].shape == (2,)
    assert state['1']['mean'].shape == (2,)
    assert state['1']['var'].shape == (2,)

  def test_shared_variables(self):
    v = nnx.Param(1)
    g = [v, v]

    state, static = nnx.graph_utils.graph_flatten(g)

    assert len(state.flat_state()) == 1

    g2 = static.merge(state)

    assert g2[0] is g2[1]

  def test_tied_weights(self):
    class Foo(nnx.Module):
      def __init__(self, *, rngs: nnx.Rngs) -> None:
        self.bar = nnx.Linear(2, 2, rngs=rngs)
        self.baz = nnx.Linear(2, 2, rngs=rngs)

        # tie the weights
        self.baz.variables.kernel = self.bar.variables.kernel

    node = Foo(rngs=nnx.Rngs(0))
    state, static = nnx.graph_utils.graph_flatten(node)

    assert len(state.flat_state()) == 3  # 2 bias + 1 kernel

    node2 = static.merge(state)

    assert node2.bar.variables.kernel is node2.baz.variables.kernel

  def test_tied_weights_example(self):
    class LinearTranspose(nnx.Module):
      def __init__(self, dout: int, din: int, *, rngs: nnx.Rngs) -> None:
        self.kernel = nnx.Param(
          nnx.initializers.lecun_normal()(rngs(), (dout, din))
        )

      def __call__(self, x):
        return x @ self.kernel.T

    class Encoder(nnx.Module):
      def __init__(self, *, rngs: nnx.Rngs) -> None:
        self.embed = nnx.Embed(10, 2, rngs=rngs)
        ...
        self.linear_out = LinearTranspose(10, 2, rngs=rngs)

        # tie the weights
        self.linear_out.variables.kernel = self.embed.variables.embedding

      def __call__(self, x):
        x = self.embed(x)
        ...
        return self.linear_out(x)

    model = Encoder(rngs=nnx.Rngs(0))
    state, static = model.split()

    assert len(state.flat_state()) == 1

    x = jax.random.randint(jax.random.key(0), (2,), 0, 10)
    y = model(x)

    assert y.shape == (2, 10)

  def test_state_variables_not_shared_with_graph(self):
    class Foo(nnx.Module):
      def __init__(self):
        self.a = nnx.Param(1)

    m = Foo()
    state, static = m.split()

    assert isinstance(m.variables.a, nnx.Param)
    assert isinstance(state.variables.a, nnx.Param)
    assert m.variables.a is not state.variables.a
    assert m.a == state.a

    m2 = static.merge(state)

    assert isinstance(m2.variables.a, nnx.Param)
    assert isinstance(state.variables.a, nnx.Param)
    assert m2.variables.a is not state.variables.a
    assert m2.a == state.a

  def test_shared_state_variables_not_shared_with_graph(self):
    class Foo(nnx.Module):
      def __init__(self):
        p = nnx.Param(1)
        self.a = p
        self.b = p

    m = Foo()
    state, static = m.split()

    assert isinstance(m.variables.a, nnx.Param)
    assert isinstance(m.variables.b, nnx.Param)
    assert isinstance(state.variables.a, nnx.Param)
    assert 'b' not in state
    assert m.variables.a is not state.variables.a
    assert m.variables.b is not state.variables.a
    assert m.a == state.a
    assert m.b == state.a

    m2 = static.merge(state)

    assert isinstance(m2.variables.a, nnx.Param)
    assert isinstance(m2.variables.b, nnx.Param)
    assert isinstance(state.variables.a, nnx.Param)
    assert m2.variables.a is not state.variables.a
    assert m2.variables.b is not state.variables.a
    assert m2.a == state.a
    assert m2.b == state.a
    assert m2.a is m2.b



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from copy import deepcopy
from typing import Any, TypeVar

import jax
import jax.numpy as jnp
import pytest

from flax.experimental import nnx

A = TypeVar('A')


class TestModule:
  def test_has_module_state(self):
    class Foo(nnx.Module):
      ...

    foo = Foo()

    assert hasattr(foo, '_module__state')

  def test_trace_level(self):
    m = nnx.Dict(a=nnx.Param(1))

    @jax.jit
    def f():
      with pytest.raises(
        nnx.TraceContextError,
        match='Cannot mutate Module from different trace level',
      ):
        m.a = 2

    f()

  def test_tree_map(self):
    m = nnx.Dict(a=nnx.Param(1))

    state, static = m.split()

    state = jax.tree_map(lambda x: x + 1, state)

  def test_split_2(self):
    m = nnx.Dict(a=nnx.Param(1))

    empty, some, static = m.split(None, ...)

    some = jax.tree_map(lambda x: x + 1, some)

  def test_split_merge(self):
    m = nnx.Dict(a=nnx.Param(1))

    @jax.jit
    def g(state: nnx.State, graphdef: nnx.GraphDef[nnx.Dict[int]]):
      m = graphdef.merge(state)
      m.a = 2
      return m.split()

    state, graphdef = g(*m.split())
    m2 = graphdef.merge(state)

    assert m2.a == 2

  def test_no_trace_level_error_on_grad(self):
    # No trace level error occurs because jax doesn't update
    # its top trace for grad.
    m = nnx.Dict(a=nnx.Param(1.0))

    @jax.grad
    def f(_):
      m.a = 2.0
      return 1.0

    f(1.0)

  def test_trace_level_error_on_nnx_grad(self):
    # error occurs because nnx updates its nnx_trace
    # in nnx.grad.
    m = nnx.Dict(a=nnx.Param(1.0))

    @nnx.grad
    def f(_):
      with pytest.raises(
        nnx.TraceContextError,
        match='Cannot mutate Module from different trace level',
      ):
        m.a = 2.0
      return 1.0

    f(m)

  def test_call(self):
    class Foo(nnx.Module):
      def __init__(self, c: float, *, rngs: nnx.Rngs):
        key = rngs.params()
        self.w = nnx.Param(jax.random.uniform(key, ()))
        self.c = c

      def __call__(self, x, *, rngs: nnx.Rngs):
        key = rngs.e()
        return self.w * x + jax.random.normal(key, ()) + self.c

    foo = Foo(c=1.0, rngs=nnx.Rngs(0))

    y = foo(x=2.0, rngs=nnx.Rngs(e=1))

    assert isinstance(y, jax.Array)

  def test_shared_module(self):
    m1 = nnx.Dict(a=nnx.Param(1), b=nnx.Param(2))
    m2 = nnx.Dict(x=m1, y=m1, z=nnx.Param(3))

    m3 = nnx.merge(m2.split())

    assert m3['x'] is m3['y']
    assert m3['x']['a'] is m3['y']['a']
    assert m3['x']['b'] is m3['y']['b']

  def test_module_graph(self):
    class Foo(nnx.Module):
      def __init__(self):
        self.a = nnx.Param(1)
        self.sub = self

    m = Foo()

    state, graphdef = m.split()
    assert len(state) == 1

    m2 = graphdef.merge(state)
    assert m2 is m2.sub

  def test_deref_through_jit(self):
    r1 = nnx.Variable(1)
    r2 = nnx.Variable(2)

    m = m0 = nnx.Dict({'a': nnx.Sequence([r1, r2]), 'b': r1})

    @jax.jit
    def f(state: nnx.State, graphdef: nnx.GraphDef[nnx.Dict[Any]]):
      m = graphdef.merge(state)

      assert m['a'][0] is m['b']
      assert m['a'][1] is not m['b']

      return m.split()

    state, graphdef = f(*m.split())
    m = graphdef.merge(state)

    assert m['a'][0] is m['b']
    assert m['a'][1] is not m['b']

    # compare with original
    assert m['a'][0] is not m0['a'][0]
    assert m['a'][1] is not m0['a'][1]
    assert m['b'] is not m0['b']

  def test_cross_barrier(self):
    m = nnx.Dict(a=nnx.Param(1))

    @jax.jit
    def g(state: nnx.State, graphdef: nnx.GraphDef[nnx.Dict[int]]):
      m = graphdef.merge(state)
      m.a += 1
      return m.split()

    state, graphdef = g(*m.split())
    m2 = graphdef.merge(state)
    assert m2 is not m
    assert m.a == 1
    assert m2.a == 2

  def test_no_rejit(self):
    n = 0
    m = nnx.Dict(a=nnx.Param(1))

    @jax.jit
    def g(state_and_def):
      nonlocal n
      n += 1
      m = nnx.merge(state_and_def)
      m.a += 1
      return m.split()

    m2 = nnx.merge(g(m.split()))

    assert n == 1
    assert m2 is not m
    assert m.a == 1
    assert m2.a == 2

    g(m.split())
    assert n == 1

    g(m2.split())
    assert n == 1

    m2.b = nnx.Param(10)
    g(m2.split())

    assert n == 2

  def test_deref_number_of_fields(self):
    r1 = nnx.Variable(1)
    r2 = nnx.Variable(2)
    v1 = 3
    m = nnx.Dict(
      {
        'a': nnx.Sequence([r1, r2, v1]),
        'b': nnx.Dict({'c': r1, 'd': r2}),
      }
    )

    p, graphdef = m.split()
    assert len(p.flat_state()) == 2
    assert len(jax.tree_util.tree_leaves(p)) == 2

  def test_deref_array_attributes_not_allowed(self):
    # test arrays are nodes
    r1 = nnx.Variable(1)
    r2 = nnx.Variable(2)
    v1 = jax.numpy.array(3)

    with pytest.raises(
      ValueError,
      match=f"Trying to assign a '{type(v1).__name__}' to the Module",
    ):
      m = nnx.Dict(
        {
          'a': nnx.Sequence([r1, r2, v1]),
          'b': nnx.Dict({'c': r1, 'd': r2}),
        }
      )

  def test_clone(self):
    m = nnx.Dict(
      a=nnx.Sequence([nnx.Param(1), nnx.Param(2), 3]),
      b=nnx.Dict(c=nnx.Param(1), d=nnx.Param(2)),
    )

    m2 = m.clone()

    assert m is not m2
    assert m2.a[0] == m2.b.c
    assert m2.a[1] == m2.b.d

    assert m.a[0] == m2.a[0]
    assert m.a[1] == m2.a[1]
    assert m.b.c == m2.b.c
    assert m.b.d == m2.b.d

  def test_sow_basic(self):
    class Foo(nnx.Module):
      def __call__(self, x):
        y = x + 1
        self.sow(nnx.Intermediate, 'y', y)
        return y

    m = Foo()
    y1 = m(2)
    y2 = m(10)

    assert y1 == 3
    assert y2 == 11
    assert m.y == (3, 11)

    intermediates = m.pop(nnx.Intermediate)

    assert isinstance(intermediates.variables.y, nnx.Intermediate)
    assert intermediates['y'] == (3, 11)

    assert hasattr(m, 'y')
    assert m.y is nnx.EMPTY

  def test_sow_existing_non_variable_field(self):
    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.y = 10

      def __call__(self, x):
        y = x + 1
        self.sow(nnx.Intermediate, 'y', y)
        return y

    m = Foo()

    with pytest.raises(ValueError, match='to be a Variable, got'):
      m(2)

  def test_sow_wrong_collection(self):
    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.y = nnx.Param(10)

      def __call__(self, x):
        y = x + 1
        self.sow(nnx.Intermediate, 'y', y)
        return y

    m = Foo()

    with pytest.raises(ValueError, match='to be of type'):
      m(2)

  def test_update_static_state(self):
    class Foo(nnx.Module):
      def add_field(self):
        self.a = 1

    m1 = Foo()
    m2 = Foo()
    m2.add_field()

    m1.update(m2)

    assert m1.a == 1

  def test_update_moduledef(self):
    class Foo(nnx.Module):
      def add_field(self):
        self.a = 1

    m1 = Foo()
    m2 = Foo()
    m2.add_field()

    m1.update(m2.get_graphdef())

    assert m1.a == 1

  def test_update_static_state_submodules(self):
    class Bar(nnx.Module):
      def __init__(self) -> None:
        self.x = 1

      def add_field(self):
        self.y = 2

    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.a = Bar()
        self.b = self.a

    m1 = Foo()
    m2 = Foo()
    m2.a.add_field()

    m1.update(m2)

    assert m1.a.x == 1
    assert m1.a.y == 2
    assert m1.b.x == 1
    assert m1.b.y == 2

  def test_update_new_submodule(self):
    class Bar(nnx.Module):
      def __init__(self) -> None:
        self.x = 1

    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.a = Bar()

      def add_module(self):
        self.b = Bar()

    m1 = Foo()
    m2 = Foo()
    m2.add_module()

    m1.update(m2)

    assert m1.a.x == 1
    assert m1.b.x == 1

  def test_update_update_submodule(self):
    class Bar(nnx.Module):
      def __init__(self) -> None:
        self.x = 1

    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.a = Bar()
        self.b = self.a

    m1 = Foo()
    m2 = Foo()
    m2.a.x = 2

    m1.update(m2)

    assert m1.a.x == 2
    assert m1.b.x == 2

  def test_update_add_shared_error(self):
    class Bar(nnx.Module):
      def __init__(self) -> None:
        self.x = 1

    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.a = Bar()
        self.b = self.a

      def add_submodule(self):
        self.c = self.a

    m1 = Foo()
    m2 = Foo()
    m2.add_submodule()

    assert hasattr(m2, 'c')

    with pytest.raises(ValueError, match='Trying to add a new node at path'):
      m1.update(m2)

  def test_update_add_shared_error_new_first(self):
    class Bar(nnx.Module):
      def __init__(self) -> None:
        self.x = 1

    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.b = Bar()
        self.c = self.b

      def add_submodule(self):
        self.a = self.b

    m1 = Foo()
    m2 = Foo()
    m2.add_submodule()

    assert hasattr(m2, 'a')

    m2 = m2.clone()  # clone to sort the fields

    with pytest.raises(ValueError, match='Trying to add a new node at path'):
      m1.update(m2)

  def test_create_abstract(self):
    linear = nnx.Linear.create_abstract(2, 3, rngs=nnx.Rngs(0))

    assert linear.kernel == jax.ShapeDtypeStruct((2, 3), jnp.float32)
    assert linear.bias == jax.ShapeDtypeStruct((3,), jnp.float32)

  def test_deepcopy(self):
    class Foo(nnx.Module):
      def __init__(self) -> None:
        self.a = nnx.Param(1)
        self.b = [1, 2, 3]
        self.c = nnx.Param(jnp.array([1.0]))
        self.self = self

    m1 = Foo()
    m2 = deepcopy(m1)

    assert m1.a == m2.a
    assert vars(m1)['a'] is not vars(m2)['a']
    assert m1.b is not m2.b
    assert m1.c is not m2.c
    assert m1.self is m1


class TestModulePytree:
  def test_tree_map(self):
    class Foo(nnx.Module, experimental_pytree=True):
      def __init__(self):
        self.node = nnx.Param(1)
        self.static = 1

    m = Foo()

    m = jax.tree_map(lambda x: x + 1, m)

    assert m.node == 2
    assert m.static == 1


class TestModuleDataclass:
  def test_basic(self):
    @nnx.dataclass
    class Foo(nnx.Module):
      a: int
      b: int = nnx.treenode_field()
      c: int = nnx.param_field()
      d: int = nnx.variable_field(nnx.BatchStat)
      e: int
      f: int

    m = Foo(
      a=1,  # static
      b=2,  # node
      c=3,  # param
      d=4,  # var
      e=5,  # static int
      f=nnx.Variable(6),  # test that we can pass in a node
    )

    state, graphdef = m.split()

    assert len(state) == 4
    assert state.variables.b == nnx.TreeNode(2)
    assert state.variables.c == nnx.Param(3)
    assert state.variables.d == nnx.BatchStat(4)
    assert state.variables.f == nnx.Variable(6)

  def test_no_override(self):
    @nnx.dataclass
    class Foo(nnx.Module):
      a: int = nnx.treenode_field()

    with pytest.raises(ValueError, match='is not compatible with return type'):
      _m = Foo(a=nnx.Param(1))

    _m = Foo(a=nnx.TreeNode(1))

  def test_context_none_after_init(self):
    @dataclasses.dataclass
    class DFoo(nnx.Module):
      din: int
      dout: int
      rngs: nnx.Rngs

      def __post_init__(self):
        self.bar = nnx.Linear(self.din, self.dout, rngs=self.rngs)

      def __call__(self, x):
        return self.bar(x)

    m = DFoo(1, 1, rngs=nnx.Rngs(0))

    assert hasattr(m, 'bar')
    assert m.rngs is None

  def test_setup_is_called(self):
    @dataclasses.dataclass
    class DFoo(nnx.Module):
      din: int
      dout: int
      rngs: nnx.Rngs

      def setup(self):
        self.bar = nnx.Linear(self.din, self.dout, rngs=self.rngs)

      def __call__(self, x):
        return self.bar(x)

    m = DFoo(1, 1, rngs=nnx.Rngs(0))

    assert hasattr(m, 'bar')
    assert m.rngs is None


class TestModuleDef:
  def test_apply(self):
    class Foo(nnx.Module):
      def __init__(self, c: float, *, rngs: nnx.Rngs):
        self.w = nnx.Param(jax.random.uniform(rngs.params(), ()))
        self.c = c

      def __call__(self, x, *, rngs: nnx.Rngs):
        key = rngs.e()
        return self.w * x + jax.random.normal(key, ()) + self.c

    rngs = nnx.Rngs(0)
    foo = Foo(c=1.0, rngs=rngs)

    states, graphdef = foo.split()

    assert isinstance(states, nnx.State)
    assert isinstance(states.variables.w, nnx.Param)
    # assert isinstance(states["c"], jax.Array)

    y, _updates = graphdef.apply(states)(x=2.0, rngs=nnx.Rngs(e=1))

    assert isinstance(y, jax.Array)

  def test_derefed_mod_apply(self):
    class Foo(nnx.Module):
      def __init__(self, c: float, *, rngs: nnx.Rngs):
        self.w = nnx.Param(
          jax.random.uniform(rngs.params(), ()),
        )
        self.c = nnx.Variable(c)

      def __call__(self, x, *, rngs: nnx.Rngs):
        key = rngs.e()
        return self.w * x + jax.random.normal(key, ()) + self.c

    foo = Foo(c=1.0, rngs=nnx.Rngs(0))

    state, graphdef = foo.split()

    assert isinstance(graphdef, nnx.GraphDef)
    assert isinstance(state, nnx.State)
    assert isinstance(state.variables.w, nnx.Param)
    assert isinstance(state.variables.c, nnx.Variable)

    y, (state, graphdef) = graphdef.apply(state)(x=2.0, rngs=nnx.Rngs(e=1))

    assert isinstance(y, jax.Array)

  def test_modules_iterator(self):
    class Foo(nnx.Module):
      def __init__(self, *, rngs: nnx.Rngs):
        self.submodules = [
          {'a': nnx.Linear(1, 1, rngs=rngs)},
          {'b': nnx.Conv(1, 1, 1, rngs=rngs)},
        ]

    module = Foo(rngs=nnx.Rngs(0))

    modules = list(module.modules())

    assert len(modules) == 3
    assert modules[0][0] == ''
    assert isinstance(modules[0][1], Foo)
    assert modules[1][0] == 'submodules/0/a'
    assert isinstance(modules[1][1], nnx.Linear)
    assert modules[2][0] == 'submodules/1/b'
    assert isinstance(modules[2][1], nnx.Conv)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax, jax.numpy as jnp

from flax.experimental import nnx


class TestMultiHeadAttention:
  def test_basic(self):
    module = nnx.MultiHeadAttention(
      num_heads=2,
      in_features=3,
      qkv_features=6,
      out_features=6,
      rngs=nnx.Rngs(0),
    )
    y = module(jnp.ones((1, 7, 3)), decode=False)
    assert y.shape == (1, 7, 6)

  def test_multihead_sow_attention_weights(self):
    class Model(nnx.Module):
      attention_kwargs: dict

      def __init__(self, attention_kwargs, rng):
        self.attention_layers = [
          nnx.MultiHeadAttention(**attention_kwargs, rngs=rng) for i in range(3)
        ]

      def __call__(self, x, sow_weights=False):
        x = self.attention_layers[0](x, sow_weights=sow_weights)
        x = self.attention_layers[1](x)
        x = self.attention_layers[2](x, sow_weights=sow_weights)
        return x

    rng = nnx.Rngs(0)
    x = jnp.ones((4, 6, 8))

    module = Model(
      dict(
        in_features=8,
        num_heads=8,
        kernel_init=nnx.initializers.ones_init(),
        bias_init=nnx.initializers.zeros_init(),
        deterministic=False,
      ),
      rng,
    )

    with nnx.flags(decode=False):
      _ = module(x, True)
    intermediates = module.pop(nnx.Intermediate)
    assert intermediates['attention_layers/0/attention_weights'][0].shape == (
      4,
      8,
      6,
      6,
    )
    assert 'attention_layers/1/attention_weights' not in intermediates
    assert intermediates['attention_layers/2/attention_weights'][0].shape == (
      4,
      8,
      6,
      6,
    )

    with nnx.flags(decode=False):
      _ = module(x)
    intermediates = module.pop(nnx.Intermediate)
    assert not intermediates  # empty

  def test_autoregressive_decode_with_x64(self):
    with jax.experimental.enable_x64():
      x = jnp.ones((1, 4, 4))
      module = nnx.MultiHeadAttention(
        in_features=4,
        num_heads=2,
        qkv_features=4,
        decode=True,
        rngs=nnx.Rngs(0),
      )
      module.init_cache(x.shape, dtype=x.dtype)
      assert module.cached_key.shape == (1, 4, 2, 2)
      assert module.cached_value.shape == (1, 4, 2, 2)

      y1 = module(x[:, :1, :])
      y2 = module(x[:, 1:2, :])

      assert y1.shape == (1, 1, 4)
      assert y2.shape == (1, 1, 4)



# Copyright 2024 The Flax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
import optax

from flax.experimental import nnx


class TestHelpers:
  def test_train_state(self):
    m = nnx.Dict(a=nnx.Param(1), b=nnx.BatchStat(2))

    params, batch_stats, graphdef = m.split(nnx.Param, nnx.BatchStat)

    state = nnx.TrainState(
      graphdef,
      params=params,
      tx=optax.sgd(1.0),
      batch_stats=nnx.TreeNode(batch_stats),
      other=nnx.Variable(100),
      int=200,
    )

    leaves = jax.tree_util.tree_leaves(state)

    assert 1 in leaves
    assert 2 in leaves
    assert 100 in leaves
    assert 200 not in leaves

  def test_train_state_methods(self):
    class Foo(nnx.Module):
      def __init__(self, *, rngs: nnx.Rngs):
        self.linear = nnx.Linear(2, 4, rngs=rngs)
        self.batch_norm = nnx.BatchNorm(4, rngs=rngs)

      def __call__(self, x: jax.Array, train: bool) -> jax.Array:
        x = self.linear(x)
        x = self.batch_norm(x, use_running_average=not train)
        return x

    module = Foo(rngs=nnx.Rngs(0))
    params, batch_stats, graphdef = module.split(nnx.Param, nnx.BatchStat)

    state = nnx.TrainState(
      graphdef,
      params=params,
      tx=optax.sgd(1.0),
      batch_stats=batch_stats,
    )

    x = jax.numpy.ones((1, 2))
    y, _updates = state.apply('params', 'batch_stats')(x, train=True)

    assert y.shape == (1, 4)

    # fake gradient
    grads = jax.tree_map(jnp.ones_like, state.params)
    # test apply_gradients
    state = state.apply_gradients(grads)
